{
  COLLAPSE (C) by Vladimir Kladov, 2005

  Purpose: make smaller applications translating some Pascal- / Assembler-made
           procedures, functions to Forth-like language and then translating it
           to compact P-code making these code smaller twice or more.

  Conditionals:

  SURE_INSTALLCOLLAPSE_CALLEDONCE - add such symbol to the project options, if
           you are sure that InstallCollapse is called only once when the
           application is started. Otherwise consiquences are bad.

}

unit Collapse;

interface

{$I KOLDEF.INC}
//{$DEFINE NOEXCEPTIONONEXIT}
{$IFNDEF noPcode}
  {$DEFINE PCode}
{$ENDIF}

uses Windows, KOL {$I CollapseUses.inc}, BaseLib;

const RetStackSize = 1024;

procedure InstallCollapse;
{* InstallCollapse must be called as soon as possible when the application is
   started, before calling any P-code procedure and never in Try-Finally/Except
   block. But the initialization of the application must already be done,
   so InstallCollapse should not be called in initialization section, but just
   before the first form creation in the project. }

procedure EmulatePCode;
{* EmulatePCode must be called from assembler code. The next byte in the code
   must be the first byte (KN) of the P-procedure. }
procedure CollapseProcTable0;
{* It is not a procedure but just a table of entries into all used in P-code
   procedures, functions, methods. }
procedure CollapseProcTable1;
procedure CollapseProcTable2;
procedure CollapseProcTable3;
{* The same as above but for Pascal-procedures having 1, 2 or 3 parameters
   passed through registers EAX, EDX, ECX. }

procedure RunPCode0;
procedure RunPCode1;
procedure RunPCode2;
procedure RunPCode3;

type
  TControl_ = object(KOL.TControl) end;
  TGraphtool = object(KOL.TGraphicTool) end;
  TTimer_ = object(KOL.TTimer) end;
  TMMTimer_ = object(KOL.TMMTimer) end;
  TMenu_ = object(KOL.TMenu) end;
  TImageList_ = object(KOL.TImageList) end;
  TThread_ = object(KOL.TThread) end;
  TTrayIcon_ = object(KOL.TTrayIcon) end;

  {$I CollapseObjects.inc}

//{$IFDEF HaltOnExit_AnyException}
  // if such symbol defined, set AppletTerminating as soon as the application
  // desides to close. If any exception will be fired after that, it will be
  // suppressed and the application will be terminated immediately without any
  // undesired messages.
var AppletTerminating: Boolean;
//{$ENDIF}

implementation

{$IFDEF COLLAPSE_DEBUG}
procedure InitCallPDebuger; forward;

var PDebuger_Module: THandle;
    PDebuger: procedure = InitCallPDebuger;
    passESI, passEDI, passEBP: DWORD;

procedure Dummy;
begin
end;

procedure DoInitCallPDebuger1;
var SL: PStrList;
    PDebugerPath: String;
    i: Integer;
    s: String;
    InitDebug: procedure( APath: PChar );
begin
  PDebugerPath := '';
  SL := NewStrList;
  TRY
    SL.LoadFromFile( GetStartDir + 'PCompiler.cfg' );
    for i := 0 to SL.Count-1 do
    begin
      s := SL.Items[ i ];
      if StrEq( Copy( s, 1, 2 ), '/D' ) then
      begin
        PDebugerPath := Trim( CopyEnd( s, 3 ) );
        break;
      end;
    end;
  FINALLY
    SL.Free;
  END;
  if DirectoryExists( PDebugerPath ) then
    PDebugerPath := IncludeTrailingPathDelimiter( PDebugerPath ) + 'PDebuger.dll';
  if not FileExists( PDebugerPath ) then
    PDebugerPath := 'PDebuger.dll';
  PDebuger_Module := LoadLibrary( PChar( PDebugerPath ) );
  if PDebuger_Module <> 0 then
  begin
    PDebuger := GetProcAddress( PDebuger_Module, 'CollapseDebug' );
    if not Assigned( PDebuger ) then
    begin
      FreeLibrary( PDebuger_Module );
      PDebuger_Module := 0;
    end;
    InitDebug := GetProcAddress( PDebuger_Module, 'InitCollapseDebug' );
    if Assigned( InitDebug ) then
      InitDebug( PChar( GetStartDir ) );
  end;
end;

procedure DoInitCallPDebuger;
begin
  DoInitCallPDebuger1;
  if PDebuger_Module = 0 then
    PDebuger := Dummy
  else
  asm
    push ESI
    push EDI
    push EBP
    mov  esi, [passESI]
    mov  edi, [passEDI]
    mov  ebp, [passEBP]
    call PDebuger
    pop  EBP
    pop  EDI
    pop  ESI
  end;
end;

procedure InitCallPDebuger;
asm
  mov [passESI], ESI
  mov [passEDI], EDI
  mov [passEBP], EBP
  call DoInitCallPDebuger;
end;

{$ENDIF COLLAPSE_DEBUG}

type
  PCrackList = ^TCrackList;
  TCrackList = object( TList )
  end;

type
  TContext = record
    ContextFlags: Pointer;

    xDr0: DWORD;
    xDr1: DWORD;
    xDr2: DWORD;
    xDr3: DWORD;
    xDr6: DWORD;
    xDr7: DWORD;

    FloatSave: TFloatingSaveArea;

    SegGs: Pointer;
    SegFs: Pointer;
    SegEs: Pointer;
    SegDs: Pointer;

    xEdi: Pointer;
    xEsi: Pointer;
    xEbx: Pointer;
    xEdx: Pointer;
    xEcx: Pointer;
    xEax: Pointer;

    xEbp: Pointer;
    xEip: Pointer;
    SegCs: Pointer;
    xEFlags: Pointer;
    xEsp: Pointer;
    SegSs: Pointer;
  end;

  TContextAMD = record
    P1Home, P2Home, P3Home, P4Home, P5Home, P6Home: Int64;
    ContextFlags: DWORD;
    MxCsr       : DWORD;

    SegCs : Word;
    SegDs : Word;
    SegEs : Word;
    SegFs : Word;
    SegGs : Word;
    SegSs : Word;
    EFlags: DWORD;

    Dr0, Dr1, Dr2, Dr3,
    Dr6   : Int64;

    Rax, Rcx, Rdx, Rbx,
    Rsp, Rbp, Rsi, Rdi,
    R8, R9, R10, R11, R12, R13, R14, R15: Int64;

    Rip: Int64;

    FltSave: array[ 0..511 ] of Byte;
    // and other states are not defined here!
  end;

procedure RunPCode0;
asm
  jmp  EmulatePCode
end;

procedure RunPCode1;
asm
  xchg [esp], eax
  push eax
  jmp  EmulatePCode
end;

procedure RunPCode2;
asm
  xchg edx, [esp]
  push eax
  push edx
  jmp  EmulatePCode
end;

procedure RunPCode3;
asm
  xchg ecx, [esp]
  push edx
  push eax
  push ecx
  jmp  EmulatePCode
end;

procedure EmulatePCode;
const MinShortCall = $C0;
asm
  {$IFDEF NT_ONLY}
  // edi -> retstack + 8
  push dword ptr [edi-4]
  sub edi, 4

  //! ret addr -> KN byte starting P-code to emulate

  xchg esi, [esp] // esi = retaddr

  cmp  byte ptr [esi-1], $E4
  je   @@enter_no_param
  cmp  byte ptr [esi-1], $E6
  jl   @@enter_1param
  je   @@enter_2params
  xchg [esp], ecx
  push edx
  push eax
  push ecx
  jmp  @@enter_no_param
@@enter_2params:
  xchg [esp], edx
  push eax
  push edx
  jmp  @@enter_no_param
@@enter_1param:
  xchg [esp], eax
  push eax
@@enter_no_param:
  POP  EAX
  STOSD         // save ESI (InstrPt)
  XCHG EAX, EBX
  STOSD         // save EBX (Result)
  XOR  EAX, EAX
  STOSD         // save 0 as RetAddr
  XCHG EAX, EBP
  STOSD         // save EBP (Self)
  {$ELSE}
  {cmp  byte ptr [esi-1], $FD
  jb   @@enter_no_param
  je   @@enter_1param
  cmp  byte ptr [esi-1], $FE
  je   @@enter_2params}
  //---   5  call RunPCodeN (,  
  // ,   Windows9x,      Athlon64):
  mov eax, RetStackSize
  call System.@GetMem
  xchg edi, eax
  stosd
  xchg eax, esi
  stosd
  xchg eax, ebx
  stosd
  xor eax, eax
  stosd
  xchg eax, ebp
  stosd
  pop  esi 
  {$ENDIF}
  //! esp -> esi saved,
  //! esi = ret-addr ->   -,
  //! eax, edx, ecx are free to use


@@MainLoop:
  {$IFDEF COLLAPSE_DEBUG}
  call dword ptr[PDebuger]
  {$ENDIF COLLAPSE_DEBUG}
  //   -
  XOR EAX, EAX
  LODSB
  TEST AL, $80
  JNZ  @@CALLS
//---------- 0xxx.xxxx -     
  TEST AL, $40
  JNZ  @@condJUMPS
  TEST AL, $20
  JNZ  @@JUMP
//----------    -16..+15:
  TEST AL, $10
  JZ   @@storeCONST
  OR   EAX, $FFFFFFF0
@@storeCONST:
  PUSH EAX
  JMP  @@MainLoop
//----------  
@@JUMP:
  TEST AL, $10
  JZ   @@jumpFWD
  OR   EAX, $FFFFFFF0
@@jumpFWD:
  MOV  AH, AL
  LODSB
@@DOJUMP:
  ADD  ESI, EAX
  JMP  @@MainLoop
//----------  
@@condJUMPS:
  MOV  CL, AL
  TEST AL, $10
  JZ   @@CondJumpFwd
  OR   EAX, $FFFFFFF0
  JMP  @@doCondJump
@@CondJumpFWD:
  AND  AL, $0F
@@doCondJump:
  MOV  AH, AL
  LODSB
  POP  EDX
  TEST CL, $20
  JZ   @@jumpIF0
//----------   <> 0
  TEST EDX, EDX
  JZ   @@MainLoop //  0,  
  ADD  ESI, EAX
  JMP  @@MainLoop
//---------- ,    0
@@jumpIF0:
  TEST EDX, EDX
  JNZ  @@MainLoop
  ADD  ESI, EAX
  JMP  @@MainLoop
//----------  
@@CALLS:
  SUB  AL, MinShortCall
  JAE  @@DOCALL
  ADD  AL, MinShortCall
  TEST AL, $30
  JNZ  @@Calls_Pascal
@@LONGCALLS:
  AND  AL, $3F
  MOV  AH, AL
  LODSB
@@DOCALL:
  MOV EAX, dword ptr [CollapseProcTable0+EAX*4]

  {$IFDEF NT_ONLY}
  {$IFDEF COLLAPSE_FASTCALLS}
  CMP byte ptr[EAX], $E8
  JNE @@noFastCalls
  CMP byte ptr[EAX+5], $E4
  JB  @@CallAsmCode
  CMP byte ptr[EAX+5], $E7
  JA  @@CallAsmCode
  ADD EAX, 6
  JMP @@CALLPFast
@@noFastCalls:
  {$ENDIF}
  {$ENDIF}


  {$IFDEF NT_ONLY}
  CMP byte ptr[EAX], $E4
  JB  @@CallAsmCode
  CMP byte ptr[EAX], $E7
  JA  @@CallAsmCode
  INC  EAX //   
  {$ELSE}
  {CMP byte ptr[EAX], $CD
  JNE @@CallAsmCode
  INC  EAX //     INT ($CD),   0..3
           // =    }
  CMP byte ptr[EAX], $E8
  JNE @@CallAsmCode
  mov edx, [eax+1]
  lea edx, [eax+5+edx]
  CMP edx, offset[RunPCode0]
  JB  @@CallAsmCode
  CMP edx, offset[RunPCode3]
  JA  @@CallAsmCode
  ADD EAX, 5
  {$ENDIF}
//  -  -
@@CALLPFast:
  XCHG EAX, ESI
  STOSD
  MOV  EAX, EBP
  STOSD
  JMP @@MainLoop
//------        ---
@@CallAsmCode:
  CALL EAX
  XCHG EBX, EAX
  JMP @@MainLoop

//------   -  1, 2, 3   
@@Calls_Pascal:
  AND AL, $3F
  MOVZX ECX, AL
  SHR ECX, 4
  AND AL, $0F
  MOV AH, AL
  LODSB
  PUSH EAX
  LOOP @@Call_P2
  CALL PASCAL1
  XCHG EBX, EAX
  JMP  @@MainLoop
@@Call_P2:
  LOOP @@Call_P3
  CALL PASCAL2
  XCHG EBX, EAX
  JMP  @@MainLoop
@@Call_P3:
  CALL PASCAL3
  XCHG EBX, EAX
  JMP  @@MainLoop
end;

procedure CollapseProcTable0;
asm
  {
         CollapseProcTable.inc file must contain up to 64K entries written
         in assembler style using DD directive, e.g.:

         DD TMyObject1.Method1
         DD AllocMem
         DD MyProc
         ...

         Create also CollapseUses.inc file and fill it with a list of units
         where all these procedures, functions and objects are from. This list
         must be started from a comma, e.g:

         , MainUnit, SpecialUnit, MyUtils
  }
  {$IFDEF PCode}
  {$I CollapseProcTable0.inc}
  {$ENDIF}
end;

procedure CollapseProcTable1;
asm
  {$IFDEF PCode}
  {$I CollapseProcTable1.inc}
  {$ENDIF}
end;

procedure CollapseProcTable2;
asm
  {$IFDEF PCode}
  {$I CollapseProcTable2.inc}
  {$ENDIF}
end;

procedure CollapseProcTable3;
asm
  {$IFDEF PCode}
  {$I CollapseProcTable3.inc}
  {$ENDIF}
end;

var Collapse_OldHandler: Pointer;

procedure Collapse_Handler; //( Except  : ^TExceptionRecord;
                            //  Frame   : ^_EXCEPTION_REGISTRATION ???;
                            //  Context : ^TContext;
                            //  Dispatch: ??? ): Integer; Cdecl;
(*
struct _EXCEPTION_REGISTRATION
{
  //    
  _EXCEPTION_REGISTRATION  *prev;
  //  ,  Runtime 
  SEHHandler               handler;
  //   ,   __try__except
  PSCOPETABLE              scopetable;
  //     try
  int                      trylevel;
  //    
  int                      _ebp;
};
*)
asm
  mov edx, [esp+12] // edx = @ Context
  mov ecx, [edx].TContext.xEip
  {$IFDEF NT_ONLY}
  cmp byte ptr [ecx], $E4
  jb  @@to_old_handler
  cmp byte ptr [ecx], $E7
  jbe @@collapse_call_P
  {$ELSE}
  {cmp byte ptr [ecx], $CD
  je  @@collapse_call_P}
  {$ENDIF}

@@to_old_handler:
  {$IFDEF NOEXCEPTIONONEXIT}
  CMP [AppletTerminated], 0
  JNZ @@to_halt
  CMP [AppletTerminating], 0
  JZ  @@to_old_2
@@to_halt:
  JMP System.@Halt
@@to_old_2:
  {$ENDIF}
  jmp dword ptr [Collapse_OldHandler]

  {$IFDEF NT_ONLY}
@@collapse_call_P:
  {$IFDEF JUSTSKIP_PRIVILEGE}
  inc dword ptr [edx].TContext.xEip
  {$ELSE}
    push ecx
    push edx
    MOV  EAX, RetStackSize
    CALL System.@GetMem
    pop  edx
    pop  ecx
    push edi
    xchg edi, eax
    inc  ecx
    {$IFNDEF NT_ONLY}
    //inc  ecx //    INT  2 
    {$ENDIF}
    mov eax, [edx].TContext.xEdi
    stosd
    xchg eax, ecx
    stosd
    lea ecx, [EmulatePCode]
    mov [edx].TContext.xEip, ecx
    mov [edx].TContext.xEdi, edi
    pop edi
  {$ENDIF}
  {$ELSE}
  inc dword ptr [edx].TContext.xEip
  {$ENDIF}
  xor eax, eax // ""
end;

procedure InstallCollapse;
asm
  xor eax, eax
  {$IFNDEF SURE_INSTALLCOLLAPSE_CALLEDONCE}
  cmp [Collapse_OldHandler], eax
  jne @@exit
  {$ENDIF}
  mov edx, fs:[eax]
  mov ecx, [edx+4]
  mov [Collapse_OldHandler], ecx
  lea ecx, [Collapse_Handler]
  mov [edx+4], ecx
@@exit:
end;

initialization

finalization


end.
