unit Compile2;

interface

uses Windows, KOL;

type
  PPcompiler = ^TPCompiler;
  TPCompiler = object( TObj )
  public
    BaselibPath, ProjectPath, MapFileName: String;
    DelphiCompilerCommand: String;
    DebugMode: Boolean;
    ProcLists: array[ 0..3 ] of PStrListEx;
    MappedUnits: PStrListEx;
    UnitsList: PStrListEx;
    ErrCount: Integer;
    MacroList: PStrListEx;
    MacroParams: PStrListEx;
    CurFilePath: String;
    CurLine: Integer;
    Conditionals: PStrList;
    ChangedIdexesOfProcedures: PStrList;
    procedure CompileProject;
    function CompileProjectOnce( UseExistingProcTable: Boolean ): Boolean;
    function CompileUnit( UnitIdx: Integer; SaveChanges: Boolean;
      var HasPCode: Boolean ): Boolean;
    procedure SortProcLists;
    function SaveProcTable( SaveConst: Boolean ): Boolean;
    function ProcTabPath( n: Integer ): String;
    function ProcConstPath: String;
    function CallDelphi: Boolean;
    procedure Error( const Msg: String );
    function PresentInMappedUnits( const s: String ): Boolean;
    function IsProcMapped( Line: Integer ): Boolean;
    function ConditionalDefined( const Def: String ): Boolean;
    destructor Destroy; virtual;
  end;

var Compiler: PPcompiler;

    RecompileAll: Boolean;
    wait_input: Boolean;
    SilentIfNoError: Boolean;
    FastCalls: Boolean;

    procedure Print( const s: String );
    procedure PrintLog( const s: String );

implementation

uses Compile3;

    procedure PrintLog( const s: String );
    begin
      LogFileOutput( GetStartDir + 'PCompiler.log', s );
    end;

    procedure Print( const s: String );
    begin
      Writeln( s );
      PrintLog( s );
    end;

function CompareCallCount(const Data: Pointer; const e1,e2 : Dword) : Integer;
var ProcList: PStrListEx;
    n1, n2: Integer;
begin
  ProcList := Data;
  n1 := ProcList.Objects[ e1 ] and $FFFFFF;
  n2 := ProcList.Objects[ e2 ] and $FFFFFF;
  if n1 > n2 then Result := -1
  else
  if n1 < n2 then Result := 1
  else
  if e1 < e2 then Result := 1
  else
  if e1 > e2 then Result := -1
  else
    Result := 0;
end;

procedure SwapStrItems(const Data : Pointer; const e1,e2 : Dword);
var ProcList: PStrListEx;
begin
  ProcList := Data;
  ProcList.Swap( e1, e2 );
end;

{ TPCompiler }

function TPCompiler.CallDelphi: Boolean;
var DelphiPath, Cmd: String;
    ShowCmd: DWORD;
begin
  Result := FALSE;
  Cmd := Trim( DelphiCompilerCommand );
  if Cmd = '' then Exit;
  if Cmd[ 1 ] = '"' then
  begin
    Cmd := CopyEnd( Cmd, 2 );
    DelphiPath := Parse( Cmd, '"' );
  end
    else
    DelphiPath := Parse( Cmd, ' '#9 );
  Cmd := Trim( Cmd );
  ShowCmd := SW_SHOW;
  if SilentIfNoError then
    ShowCmd := SW_HIDE;
  Result := ExecuteWait( DelphiPath, Cmd, ProjectPath, ShowCmd, INFINITE, nil );
end;

procedure TPCompiler.CompileProject;
var j, j0: Integer;
    s: String;
    SL: PStrList;
    procedure HandleParams( stage: Integer );
    var i: Integer;
        u: String;
    begin
      for i := 0 to SL.Count-1 do
      begin
        s := SL.Items[ i ];
        u := UpperCase( s );
        if stage > 0 then
          if (s = '') or (s[1] = '#') then continue;
        if StrIsStartingFrom( PChar( u ), '/A' ) then
        begin
          RecompileAll := TRUE;
        end
          else
        if StrIsStartingFrom( PChar( u ), '/B' ) then
        begin
          BaselibPath := CopyEnd( s, 3 );
          if (pos( ':', BaseLibPath ) <= 0) and (Stage > 0) and (ProjectPath <> '') then
            BaseLibPath := IncludeTrailingPathDelimiter( ProjectPath ) + BaseLibPath;
          if not FileExists( BaseLibPath ) and
             DirectoryExists( BaseLibPath ) then
             BaseLibPath := IncludeTrailingPathDelimiter( BaseLibPath )
               + 'baselib.pas'
          else if not FileExists( BaseLibPath ) then
          begin
            BaseLibPath := IncludeTrailingPathDelimiter( ProjectPath ) + BaseLibPath;
            if not FileExists( BaseLibPath ) and
               DirectoryExists( BaseLibPath ) then
               BaseLibPath := IncludeTrailingPathDelimiter( BaseLibPath )
                 + 'baselib.pas'
          end;
        end
          else
        if StrIsStartingFrom( PChar( u ), '/D' ) then
        begin
          DebugMode := TRUE;
        end
          else
        if StrIsStartingFrom( PChar( u ), '/S' ) then
        begin
          SilentIfNoError := TRUE;
        end
          else
        if StrIsStartingFrom( PChar( u ), '/C' ) then
        begin
          DelphiCompilerCommand := CopyEnd( u, 3 );
        end
          else
        if StrIsStartingFrom( PChar( u ), '/M' ) then
        begin
          MapFileName := CopyEnd( u, 3 );
        end
          else
        if StrIsStartingFrom( PChar( u ), '/F' ) then
        begin
          FastCalls := TRUE;
        end
          else
        begin
          if Stage <> 0 then continue;
          ProjectPath := s;
          if StrIn( ExtractFileExt( s ), [ '.exe', '.dll', '.pas', '.inc', '.dpr' ] ) then
            ProjectPath := ExtractfilePath( s )
            else
          if DirectoryExists( ProjectPath ) then
            else
          if not FileExists( ProjectPath ) then
          begin
            Print( 'File or Directory not found: ' + ProjectPath );
            wait_input := TRUE;
          end;
        end;
      end;
    end;
var L: PList;
    n: Integer;
    DL: PDirList;
begin
  ProjectPath := '';
  SL := NewStrList;
  ProcLists[ 0 ] := NewStrListEx;
  ProcLists[ 1 ] := NewStrListEx;
  ProcLists[ 2 ] := NewStrListEx;
  ProcLists[ 3 ] := NewStrListEx;
  ChangedIdexesOfProcedures := NewStrList;
  TRY
    s := '';
    for j := 1 to ParamCount do
    begin
      s := s + ' ' + ParamStr( j );
      SL.Add( ParamStr( j ) );
    end;
    Print( 'params: ' + s );
    HandleParams( 0 );
    if FileExists( IncludeTrailingPathDelimiter( ProjectPath ) + 'Pcompiler.cfg' ) then
    begin
      SL.LoadFromFile( IncludeTrailingPathDelimiter( ProjectPath ) + 'Pcompiler.cfg' );
      HandleParams( 1 );
    end;
    if ProjectPath = '' then
    begin
      Print( 'source files/directory not defined ' );
      wait_input := TRUE;
    end;
    if not FileExists( BaseLibPath ) then
      BaseLibPath := 'baselib.pas';
    if not FileExists( BaseLibPath ) then
    begin
      Print( 'baselib.pas not found! (' + BaseLibPath + ')' );
      BaseLibPath := '';
      wait_input := TRUE;
    end;
    if MapFileName = '' then
    begin
      Print( 'Map file not defined' );
      wait_input := TRUE;
    end
      else
    begin
      MapFileName := IncludeTrailingPathDelimiter( ProjectPath ) +
        MapFileName;
    end;
  FINALLY
    SL.Free;
  END;
  if wait_input then Exit;
  if not FileExists( MapFileName ) then
  begin
    //   -    
    Print( 'Creating map file ' + MapFileName );
    if DelphiCompilerCommand = '' then
    begin
      Print( 'Delphi compiler command is not defined (/C), ' +
               'compile project by yourself' );
      Exit;
    end;
    if not CallDelphi
    then begin
           Print( 'Delphi could not be called to recompile project, recompile it ' +
                    'by yourself' ); wait_input := TRUE;
           Exit;
         end;
    if not FileExists( MapFileName ) then
    begin
      Print( 'Map-file not found, second pass is not possible!' );
      wait_input := TRUE;
      Exit;
    end;
  end;
  UnitsList := NewStrListEx;
  UnitsList.Add( BaseLibPath );
  //  map-
  MappedUnits := NewStrListEx;
  TRY
    SL := NewStrList;
    TRY
      SL.LoadFromFile( MapFileName );
      j := -1;
      while j < SL.Count-1 do
      begin
        Inc( j );
        s := SL.Items[ j ];
        if StrIsStartingFrom( PChar( s ), 'Line numbers for' ) then
        begin
          Parse( s, '(' );
          s := Trim( Parse( s, ')' ) );
          L := NewList;
          MappedUnits.AddObject( s, DWORD( L ) );
          Inc( j );
          s := SL.Items[ j ];
          if Trim( s ) = '' then Inc( j );
          Dec( j );
          //  ,      MappedUnits.Items[ ],
          //     ,  
          //  - ,     :
          while j < SL.Count-1 do
          begin
            Inc( j );
            s := Trim( SL.Items[ j ] );
            if s = '' then break;
            while s <> '' do
            begin
              n := Str2Int( s );
              Parse( s, ' ' ); s := Trim( s );
              Parse( s, ' ' ); s := Trim( s );
              L.Add( Pointer( n ) );
            end;
          end;
        end;
      end;
    FINALLY
      SL.Free;
    END;

    DL := NewDirListEx( ProjectPath, '*.pas;*.inc', FILE_ATTRIBUTE_NORMAL );
    TRY
      for j := 0 to DL.Count-1 do
      begin
        s := DL.Names[ j ];
        if (UnitsList.IndexOf_NoCase( DL.Path + s ) < 0) and
           not( (Length( s ) = 22 ) and
                StrIsStartingFrom( PChar( UpperCase( s ) ), 'COLLAPSEPROCTABLE' )
                and StrEq(CopyEnd( s, 19 ), '.INC') ) and
           PresentInMappedUnits( DL.Names[ j ] ) then
        UnitsList.Add( DL.Path + DL.Names[ j ] );
      end;
    FINALLY
      DL.Free;
    END;

    //     ,   
    //       -,   
    //  
    Print( 'Stage #1' );
    if not CompileProjectOnce( {UseExistingProcTable=} FALSE ) then Exit;
    //         
    SortProcLists;
    // ,   ,   :
    for j0 := 0 to 3 do
      for j := ProcLists[ j0 ].Count-1 downto 0 do
      begin
        if ProcLists[ j0 ].Objects[ j ] and $FFFFFF = 0 then
        begin
          if (j > $43) or (j0 > 0) then
            ProcLists[ j0 ].Delete( j )
          else
            ProcLists[ j0 ].Items[ j ] := '0';
        end;
      end;
    //  RETURN    $C3 - $C0,   
    // :
    for j := 0 to ProcLists[ 0 ].Count-1 do
    begin
      if StrEq( ProcLists[ 0 ].Items[ j ], 'return' ) then
      begin
        if j <> 3 then
        begin
          while ProcLists[ 0 ].Count < 3 do
            ProcLists[ 0 ].AddObject( '0', 0 );
          ProcLists[ 0 ].Move( j, 3 );
        end;
        break;
      end;
    end;
    Print( 'Stage #2' );
    //    CollapseProcList.inc :
    SaveProcTable( {SaveConst=} TRUE );
    //  -   Delphi-
    CompileProjectOnce( {UseExistingProcTable=} TRUE );
    Print( 'All passes done OK' )
  FINALLY
    for j := 0 to MappedUnits.Count-1 do
      PObj( MappedUnits.Objects[ j ] ).Free;
    MappedUnits.Free;
    ProcLists[ 0 ].Free;
    ProcLists[ 1 ].Free;
    ProcLists[ 2 ].Free;
    ProcLists[ 3 ].Free;
    if (ErrCount = 0) and SilentIfNoError then
      wait_input := FALSE;
    for j := 0 to UnitsList.Count-1 do
      PObj( UnitsList.Objects[ j ] ).Free;
    UnitsList.Free;
    ChangedIdexesOfProcedures.Free;
  END;
end;

function TPCompiler.CompileProjectOnce(UseExistingProcTable: Boolean): Boolean;
var j, k: Integer;
    Mcount: Integer;
    HasPCode: Boolean;
begin
  Result := FALSE;
  MacroList := NewStrListEx;
  MacroParams := NewStrListEx;
  Mcount := 0; //    BaseLib.pas
  TRY
    j := 0;
    while j < UnitsList.Count do
    begin
      //      
      if not CompileUnit( {UnitIdx=    } j,
                          {SaveChanges=} UseExistingProcTable,
                          HasPCode ) then
        Exit; // ,    
      if j = 0 then
      begin
        Mcount := MacroList.Count;
        inc( j );
      end
      else
      begin //    ,  BaseLib.pas,  
            //  :
        for k := MacroList.Count-1 downto Mcount do
        begin
          PObj( MacroList.Objects[ k ] ).Free;
          MacroList.Delete( k );
        end;
        if not HasPCode then
        begin
          PObj( UnitsList.Objects[ j ] ).Free;
          UnitsList.Delete( j );
        end
          else
        begin
          inc( j );
        end;
      end;
    end;
    Result := TRUE;
  FINALLY
    for k := 0 to MacroList.Count-1 do
      PObj( MacroList.Objects[ k ] ).Free;
    MacroList.Free;
    MacroParams.Free;
  END;
end;

function TPCompiler.CompileUnit(UnitIdx: Integer;
  SaveChanges: Boolean; var HasPCode: Boolean): Boolean;
var U: PUnit;
begin
  U := Pointer( UnitsList.Objects[ UnitIdx ] );
  if U = nil then
  begin
    new( U, Create );
    UnitsList.Objects[ UnitIdx ] := DWORD( U );
    U.UnitIdx := UnitIdx;
  end
    else
  begin
    if not RecompileAll then
    if SaveChanges and
       (UnitIdx > 0) and not U.HasChangedCode and not U.HasChangedProcIndexes then
    begin
      Result := TRUE; //     , 
      //   ,      
      Exit;
    end;
  end;
  HasPCode := FALSE;
  Result := U.Compile( SaveChanges, HasPCode );
end;

function TPCompiler.ConditionalDefined(const Def: String): Boolean;
var CC: String;
    s: String;
    i: Integer;
begin
  if Conditionals = nil then
  begin
    Conditionals := NewStrList;
    CC := DelphiCompilerCommand;
    while CC <> '' do
    begin
      if CC[ 1 ] = '"' then
      begin
        CC := CopyEnd( CC, 2 );
        s := Trim( Parse( CC, '"' ) );
        CC := Trim( CC );
      end
        else
      begin
        s := Trim( Parse( CC, ' '#9 ) );
        CC := Trim( CC );
      end;
      if (s <> '') and StrEq( copy( s, 1, 2 ), '-D' ) then
        Conditionals.Add( CopyEnd( s, 3 ) );
    end;
  end;
  for i := 0 to Conditionals.Count-1 do
  begin
    if StrEq( Conditionals.Items[ i ], Def ) then
    begin
      Result := TRUE;
      Exit;
    end;
  end;
  Result := FALSE;
end;

destructor TPCompiler.Destroy;
begin
  BaselibPath := '';
  ProjectPath := '';
  DelphiCompilerCommand := '';
  MapFileName := '';
  CurFilePath := '';
  Conditionals.Free;
  inherited;
end;

procedure TPCompiler.Error(const Msg: String);
begin
  Print( ExtractFileName( CurFilePath ) + Format( '%6d', [ CurLine+1 ] ) +
           ': ' + Msg );
  Inc( ErrCount );
  wait_input := TRUE;
end;

function TPCompiler.IsProcMapped(Line: Integer): Boolean;
var UnitName: String;
    i: Integer;
    L: PList;
begin
  UnitName := ExtractFileName( CurFilePath );
  for i := 0 to MappedUnits.Count-1 do
  begin
    if StrEq( MappedUnits.Items[ i ], UnitName ) then
    begin
      L := Pointer( MappedUnits.Objects[ i ] );
      Result := L.IndexOf( Pointer( Line ) ) >= 0;
      Exit;
    end;
  end;
  Result := FALSE;
end;

function TPCompiler.PresentInMappedUnits(const s: String): Boolean;
var i: Integer;
begin
  Result := TRUE;
  for i := 0 to MappedUnits.Count-1 do
  begin
    if StrEq( s, ExtractFileName( MappedUnits.Items[ i ] ) ) then
      Exit;
  end;
  Result := FALSE;
end;

function TPCompiler.ProcConstPath: String;
begin
  Result := IncludeTrailingPathDelimiter( ProjectPath ) + 'CollapseProcConst.inc';
end;

function TPCompiler.ProcTabPath(n: Integer): String;
begin
  Result := IncludeTrailingPathDelimiter( ProjectPath ) + 'CollapseProcTable' +
    Int2Str( n ) + '.inc';
end;

function TPCompiler.SaveProcTable( SaveConst: Boolean ): Boolean;
const Min1ByteCall = $C0;
      Min2BytesCall = $8000;
var j0, j, k, n, n1: Integer;
    SL, SL2, SL3: PStrList;
    s, s1: String;
begin
  Result := FALSE;
  SL3 := NewStrList;
  for j0 := 0 to 3 do
  begin
    SL2 := NewStrList;
    for j := 0 to ProcLists[ j0 ].Count-1 do
    begin
      s := ProcLists[ j0 ].Items[ j ];

      if j0 = 0 then
        k := j+Min1ByteCall
      else
        k := 256;

      if k <= 255 then
      begin
        while StrReplace( s, '.', '_' ) do;
        s := Copy( s, 1, 30 );
        SL3.Add( '  n_' + s + ' = $' + Int2Hex( k, 2 ) + ';' );
        s := ProcLists[ j0 ].Items[ j ];
        s := '{ $' + Int2Hex( k, 2 ) + ' } DD ' + s;
      end
      else
      begin
        k := Min2BytesCall + j + (j0 shl 12);
        while StrReplace( s, '.', '_' ) do;
        s := Copy( s, 1, 30 );
        SL3.Add( '  n_' + s + ' = $' +
          Int2Hex( (k shr 8) or (k and $FF) shl 8, 4 ) + ';' );
        s := ProcLists[ j0 ].Items[ j ];
        s := '{ $' + Int2Hex( k shr 8, 2 ) + ' $' + Int2Hex( k and $FF, 2 ) +
          ' } DD ' + s;
      end;
      SL2.Add( s );
    end;
    SL := NewStrList;
    TRY
      SL.LoadFromFile( ProcTabPath( j0 ) );
      if (SL.Text <> SL2.Text) or (SL2.Count = 0) then
      begin
        //    ChangedIdexesOfProcedures  ,
        //       ,   
        //    :
        for j := 0 to ProcLists[ j0 ].Count-1 do
        begin
          s := ProcLists[ j0 ].Items[ j ];
          if (j <= 63) and (j0 = 0) then n := 1
          else if j <= 4095 then n := 2
          else n := 3;
          n1 := -1;
          for k := 0 to SL.Count-1 do
          begin
            s1 := SL.Items[ k ];
            Parse( s1, '}' ); s1 := Trim( s1 );
            if StrEq( s, s1 ) then
            begin
              if (k <= 63) and (j0 = 0) then n1 := 1
              else if k <= 4095 then n1 := 2
              else n1 := 3;
              if n <> n1 then
                ChangedIdexesOfProcedures.Add( s );
              break;
            end;
          end;
          if n1 < 0 then
            ChangedIdexesOfProcedures.Add( s );
        end;
        //     
        SL2.SaveToFile( ProcTabPath( j0 ) );
        Result := TRUE;
      end;
    FINALLY
      SL.Free;
      SL2.Free;
    END;
  end;
  if SaveConst then
  begin
    SL := NewStrList;
    TRY
      if SL3.Count > 0 then
        SL3.Insert( 0, 'const' );
      SL.LoadFromFile( ProcConstPath );
      if SL.Text <> SL3.Text then
      begin
        SL3.SaveToFile( ProcConstPath );
        Result := TRUE;
      end;
    FINALLY
      SL.Free;
    END;
  end;
end;

procedure TPCompiler.SortProcLists;
var j, k, n: Integer;
    Temp: PStrListEx;
begin
  Temp := NewStrListEx;
  for j := 0 to 3 do
  begin
    if ProcLists[ j ].Count > 1 then
      SortData( ProcLists[ j ], ProcLists[ j ].Count-1, CompareCallCount, SwapStrItems );
    if j = 0 then
      n := 63
    else
      n := 4095;
    for k := 0 to Min( n, ProcLists[ j ].Count-1 ) do
      Temp.AddObject( ProcLists[ j ].Items[ k ], ProcLists[ j ].Objects[ k ] );
    Temp.Sort( {CaseSensitive=} FALSE );
    for k := 0 to Temp.Count-1 do
    begin
      ProcLists[ j ].Items[ k ] := Temp.Items[ k ];
      ProcLists[ j ].Objects[ k ] := Temp.Objects[ k ];
    end;
    Temp.Clear;
    if j = 0 then
    begin
      for k := 64 to Min( 4095, ProcLists[ j ].Count-1 ) do
        Temp.AddObject( ProcLists[ j ].Items[ k ], ProcLists[ j ].Objects[ k ] );
      Temp.Sort( {CaseSensitive=} FALSE );
      for k := 0 to Temp.Count-1 do
      begin
        ProcLists[ j ].Items[ k+64 ] := Temp.Items[ k ];
        ProcLists[ j ].Objects[ k+64 ] := Temp.Objects[ k ];
      end;
      Temp.Clear;
    end;
    for k := 4096 to ProcLists[ j ].Count-1 do
      Temp.AddObject( ProcLists[ j ].Items[ k ], ProcLists[ j ].Objects[ k ] );
    Temp.Sort( {CaseSensitive=} FALSE );
    for k := 0 to Temp.Count-1 do
    begin
      ProcLists[ j ].Items[ k+4096 ] := Temp.Items[ k ];
      ProcLists[ j ].Objects[ k+4096 ] := Temp.Objects[ k ];
    end;
    Temp.Clear;
  end;
end;

end.
