unit Compile3;

interface

uses Windows, KOL;

type
  TJumpCond = ( jAlways, jIf0, jIf1 );

  PUnit = ^TUnit;
  TUnit = object( TObj )
  public
    UnitIdx: Integer;
    UnitDebugInfo: PStrList;
    AsmCode, EndCode: PStrListEx;
    Asm_firstline: Integer;
    UnitProcUsed: PStrList;
    HasChangedCode: Boolean;
    procedure Init; virtual;
    destructor Destroy; virtual;
    function Compile( SaveChanges: Boolean; var HasPCode: Boolean ): Boolean;
    function CompilePsource( SL: PStrListEx; Idx, ToIdx: Integer;
      UseExistingProcTable: Boolean ): Boolean;
    procedure AddComment_CompiledLine( var s: String; line: Integer );
    procedure AddLineToTheEnd( const s: String );
    procedure AddDebugInfo( Line, OffsetInLine, OffsetInProc: Integer );
    function HasChangedProcIndexes: Boolean;
  public
    NeedNewAsmLine: Boolean;
  end;

implementation

uses Compile2, Compile4, SUtils;

{ TUnit }

procedure TUnit.AddComment_CompiledLine(var s: String; line: Integer);
var i: Integer;
begin
  i := pos( '//--by PCompiler:', s );
  if i > 0 then
    s := Copy( s, 1, i-1 );
  s := TrimRight( s ) + ' //--by PCompiler:line#' + Int2Str( line );
end;

procedure TUnit.AddDebugInfo(Line, OffsetInLine, OffsetInProc: Integer);
var s: String;
    i, L: Integer;
begin
  //if MacroLevel > 0 then Exit;
  if (UnitDebugInfo.Count > 0) and
     (Str2Int( UnitDebugInfo.Last ) = Line) then
  begin
    s := Trim( UnitDebugInfo.Last );
    L := 0;
    for i := Length( s ) downto 1 do
      if s[ i ] = ':' then
      begin
        L := Str2Int( CopyEnd( s, i+1 ) );
        break;
      end;
    UnitDebugInfo.Items[ UnitDebugInfo.Count-1 ] :=
      s + '/' + Int2Str( OffsetInProc - L ) + ' ' +
      Int2Str( OffsetInLine ) + ':' + Int2Str( OffsetInProc );
  end
    else
  begin
    if (UnitDebugInfo.Count > 0) and
       (Str2Int( UnitDebugInfo.Last ) = Line-1) then
    begin
      s := Trim( UnitDebugInfo.Last );
      for i := Length( s ) downto 1 do
      begin
        if s[ i ] = ':' then
        begin
          L := Str2Int( CopyEnd( s, i+1 ) );
          UnitDebugInfo.Items[ UnitDebugInfo.Count-1 ] :=
            s + '/' + Int2Str( OffsetInProc - L );
          break;
        end;
      end;
    end;
    s := Format( '%.05d|%.05d', [ Line,
         Asm_firstline + AsmCode.Count + Integer( NeedNewAsmLine ) ] );
    {if MacroLevel > 0 then
    begin
      if StrEq( ExtractfileName( Compiler.CurFilePath ),
                ExtractFileName( Compiler.BaseLibPath ) ) then
        s := s + '*'
      else
        s := s + '+';
    end;}
    UnitDebugInfo.Add( s + ' ' +
      Int2Str( OffsetInLine ) + ':' + Int2Str( OffsetInProc ) );
  end;
end;

procedure TUnit.AddLineToTheEnd(const s: String);
begin
  if EndCode.IndexOf( s ) < 0 then
    EndCode.Add( s );
end;

function TUnit.Compile(SaveChanges: Boolean;
  var HasPCode: Boolean): Boolean;
var SL: PStrListEx;
    SL0: PStrList;
    i, j, k, n: Integer;
    s: String;
    Filename: String;
begin
  Filename := Compiler.UnitsList.Items[ UnitIdx ];
  PrintLog( 'Compiling unit ' + Filename );
  Compiler.CurFilePath := Filename;
  SL := NewStrListEx;
  SL0 := NewStrList;
  SL.LoadFromFile( Filename );
  for i := 0 to SL.Count-1 do
    SL.Objects[ i ] := i; //  Objects[ ]    
  SL0.Assign( SL );       // SL0 =   (   )
  UnitDebugInfo := NewStrList;
  TRY
    i := 0;
    while i < SL.Count do
    begin
      s := UpperCase( Trim( SL.Items[ i ] ) );
      if s = '{$IFDEF PSOURCE}' then
      begin
        for j := i+1 to SL.Count-2 do
        begin
          s := UpperCase( Trim( SL.Items[ j ] ) );
          if s = '{$ENDIF PSOURCE}' then
          begin
            s := UpperCase( Trim( SL.Items[ j+1 ] ) );
            if s = '{$ELSE OLDCODE}' then //  ,   
              HasChangedCode := TRUE;     // :  
                                          //       !
            HasPCode := TRUE;
            AsmCode := NewStrListEx;
            EndCode := NewStrListEx;
            TRY
                k := j+1;
                while TRUE do
                begin
                  s := UpperCase( Trim( SL.Items[ k ] ) );
                  if (Copy( s, 1, 2 ) = '{$') and
                     not StrIn(Copy( s, 1, 4 ), [ '{$R ', '{$I ' ]) then
                  begin
                    Asm_firstline := j+1;
                    break;
                  end;
                  SL.Delete( k );
                end;
              if CompilePsource( SL, i+1, j-1, SaveChanges ) then
              begin
                if Compiler.ErrCount > 0 then break;
                k := j+1;
                while TRUE do
                begin
                  s := UpperCase( Trim( SL.Items[ k ] ) );
                  if Copy( s, 1, 2 ) = '{$' then
                  begin
                    s :=  SL.Items[ i+1 ];
                    AddComment_CompiledLine( s, j+3 );
                    SL.Items[ i+1 ] := s;
                    for n := AsmCode.Count-1 downto 0 do
                      SL.Insert( j+1, AsmCode.Items[ n ] );
                    break;
                  end;
                  SL.Delete( k );
                end;
              end
                else
              begin
                s := AsmCode.Items[ 0 ];
                while StrReplace( s, #9, ' ' ) do;
                if (s <> '') and (pos( ' PROC ', s ) > 0) then
                begin
                  AsmCode.Insert( 0, Filename );
                  s := Trim( s );
                  s := Parse( s, ' :' );
                  Compiler.MacroList.AddObject( s, DWORD( AsmCode ) );
                  AsmCode := nil;
                end;
              end;
            FINALLY
              AsmCode.Free;
              EndCode.Free;
            END;
            i := j;
            break;
          end;
        end;
      end;
      if Compiler.ErrCount > 0 then break;
      Inc( i );
    end;
    if SaveChanges then
    begin
      if UnitDebugInfo.Count = 0 then
        DeleteFile( PChar( Filename + '.debug' ) )
      else
        UnitDebugInfo.SaveToFile( Filename + '.debug' );
      if Compiler.ErrCount = 0 then
      if SL.Text <> SL0.Text then
      begin
        Print( Filename + ' changed - rewritten' );
        wait_input := TRUE;
        //SL0.SaveToFile( Filename + '.old' );
        SL.SaveToFile( Filename );
      end;
    end;
    Result := Compiler.ErrCount = 0;
  FINALLY
    SL.Free;
    SL0.Free;
    UnitDebugInfo.Free;
  END;
end;

function TUnit.CompilePsource(SL: PStrListEx; Idx, ToIdx: Integer;
  UseExistingProcTable: Boolean): Boolean;
var s, key: String;
    i, n: Integer;
    proc_name: String;
    nPascalParams: Integer;
    Proc: PProc;
begin
  Result := FALSE;
  s := SL.Items[ Idx ];
  i := pos( '//', s );
  if i > 0 then
    s := Copy( s, 1, i-1 );
  s := UpperCase( Trim( s ) );
  proc_name := '';
  key := Parse( s, ' :' );
  if (key <> 'PROC') and not StrIsStartingFrom( PChar( key ), 'PROC(' ) then
  begin
    s := Trim( s );
    proc_name := key;
    key := Trim( Parse( s, ' ' ) );
    if not IsIdentifier( key ) then
      Exit;
    //i := -1;
    for n := 0 to 3 do
    begin
      i := Compiler.ProcLists[ n ].IndexOf_NoCase( proc_name );
      if i >= 0 then
      begin
        Compiler.ProcLists[ n ].Objects[ i ] :=
          Compiler.ProcLists[ n ].Objects[ i ] or (1 shl 23);
        break;
      end;
    end;
  end;
  Compiler.CurLine := SL.Objects[ Idx ];
  if (key <> 'PROC') and not StrIsStartingFrom( PChar( key ), 'PROC(' ) then
    Compiler.Error( 'PROC not found' );
  s := Trim( s );
  if StrEq( s, 'INLINE' ) {or StrEq( s, 'SMART' ) and AllSmartInline} then
  begin
    AsmCode.Clear;
    for i := Idx to ToIdx do
      AsmCode.AddObject( SL.Items[ i ], SL.Objects[ i ] );
    Exit;
  end;
  //Result := TRUE;
  nPascalParams := 0;
  Parse( key, '(' );
  if key <> '' then
    nPascalParams := Str2Int( key );
  new( Proc, Create );
  Proc.Step := 1;
  TRY
    while Proc.Step <= 2 do
    begin
      if Proc.Step >= 2 then
      begin
        AsmCode.Add( '{$I CollapseProcConst.inc}' );
        AsmCode.Add( 'asm' );
        if FastCalls then
          AsmCode.Add( '  call EmulatePCodd' );
      end;
      Proc.Offset := 0; //    -   
      Proc.IFindex := 0;
      Proc.UnitObj := @ Self;
      Proc.LineIdx := Idx - 1;
      Proc.ToLineIdx := ToIdx;
      Proc.Lines := SL;
      Proc.CompileLines( {IsInlineMacro=} FALSE, nPascalParams );
      if Proc.Step >= 2 then
      begin
        AsmCode.Add( 'end;' );
        for i := 0 to EndCode.Count-1 do
          AsmCode.Add( EndCode.Items[ i ] );
      end;
      if Compiler.ErrCount > 0 then break;
      if not UseExistingProcTable then break;
      Inc( Proc.Step );
    end;
  FINALLY
    Proc.Free;
  END;
  Result := Compiler.ErrCount = 0;
end;

destructor TUnit.Destroy;
begin
  UnitProcUsed.Free;
  inherited;
end;

function TUnit.HasChangedProcIndexes: Boolean;
var i: Integer;
    s: String;
begin
  for i := 0 to UnitProcUsed.Count-1 do
  begin
    s := UnitProcUsed.Items[ i ];
    if Compiler.ChangedIdexesOfProcedures.IndexOf_NoCase( s ) >= 0 then
    begin
      Result := TRUE;
      Exit;
    end;
  end;
  Result := FALSE;
end;

procedure TUnit.Init;
begin
  inherited;
  UnitProcUsed := NewStrList;
end;

end.
