unit Compile4;

interface

uses Windows, KOL, Compile3;

type
  PProc = ^TProc;
  TProc = object( TObj )
  public
    UnitObj: PUnit;
    InComment: Boolean;
    Lines: PStrListEx;
    LineIdx, ToLineIdx: Integer;
    EOF: Boolean;
    LineBuffer: String;
    CurLine: Integer;
    ProcMapped: Boolean;
    MacroLevel: Integer;
    SkipLevel: Integer;
    IFLevels: PStrListEx;
    IFIndex: Integer;
    Labels: PStrListEx;
    Step: Integer;
    Offset: Integer;
    procedure Init; virtual;
    destructor Destroy; virtual;
    procedure CompileLines( IsInlineMacro: Boolean; nPascalParams: Integer );
    procedure NextLine;
    procedure SkipSp( var spos: PChar );
    function SkipBrackets( var spos: PChar ): String;
    function SkipWord( var spos: PChar; const Delimiters: String;
             IncludingDelimiter: Boolean ): String;
    function SkipString( var spos: PChar ): String;
    procedure CompileBytes( const w: String );
    procedure CompileChars( w: String );
    procedure OutByte( const s: String );
    procedure OutCode( const s: String; Len: Integer );
    procedure OutWord( const s: String );
    procedure OutDWord( const s: String );
    procedure OutQWord( const s: String );
    procedure OutLabel( const s: String );
    procedure OutGoto( var w: String; cond: TJumpCond );
    procedure OutJump( ToOffset: Integer; cond: TJumpCond );
    procedure CompileIF( var w: String );
    procedure CompileElse;
    procedure CompileElseif( var w: String );
    procedure CompileEndif;
    function CompileCallMacro( var w: String; var spos: PChar ): Boolean;
    procedure CompileCallProc( var w: String; var spos: PChar );
    function Expression( var spos: PChar; const op1: String ): Integer;
    function Operand( var spos: PChar ): Integer;
    function Number( const s: String ): Integer;
    function HexNumber( const s: String ): Integer;
    function Priority( const op: String ): Integer;
    function Operation( const op: String; opd1, opd2: Integer ): Integer;
    procedure AddDebugInfo( Line, OffsetInLine, OffsetInProc: Integer );
    procedure NewAsmLine;
    function FindLabel( const s: String; DfltValue: Integer ): Integer;
    procedure UpdateLabel( const s: String; NewValue: Integer );
  end;

implementation

uses Compile2, SUtils;

function AlterCond( cond: TJumpCond ): TJumpCond;
begin
  Result := cond;
  CASE Result OF
  jIf0: Result := jIf1;
  jIf1: Result := jIf0;
  END;
end;

{ TProc }

procedure TProc.AddDebugInfo(Line, OffsetInLine, OffsetInProc: Integer);
begin
  if MacroLevel > 0 then Exit;
  UnitObj.AddDebugInfo( Line, OffsetInLine, OffsetInProc );
end;

procedure TProc.CompileBytes(const w: String);
var n, i: Integer;
begin
  if SkipLevel > 1 then Exit;
  n := 1;
  for i := 2 to Length( w ) do
    if w[ i ] = '#' then inc( n );
  CASE n OF
  1: OutByte( CopyEnd( w, 2 ) );
  2: OutWord( CopyEnd( w, 3 ) );
  4: OutDWord( CopyEnd( w, 5 ) );
  8: OutQWord( CopyEnd( w, 9 ) );
  else
    OutByte( '(' + CopyEnd( w, n+1 ) + ') and 255' );
    for i := 2 to n do
      OutByte( '((' + CopyEnd( w, n+1 ) + ') shr ' +
        Int2Str( ( i-1 ) * 8 ) +') and 255' );
  END;
end;

function TProc.CompileCallMacro(var w: String; var spos: PChar): Boolean;
  procedure DoCallMacro( Macro: PStrListEx );
  var SaveSL: PStrListEx;
      SavePath, SaveLineBuffer: String;
      SaveCurLine, SaveLineIdx, SaveToLineIdx: Integer;
      Save_spos_offset: Integer;
      SaveMacroLevel: Integer;
  begin
    SaveSL := Lines;
    SavePath := Compiler.CurFilePath;
    SaveCurLine := CurLine;
    SaveLineIdx := LineIdx;
    SaveToLineIdx := ToLineIdx;
    Save_spos_offset := DWORD( spos ) - DWORD( @ LineBuffer[ 1 ] );
    SaveLineBuffer := LineBuffer;
    SaveMacroLevel := MacroLevel;
    TRY
      Lines := Macro;
      Compiler.CurFilePath := Compiler.CurFilePath + ':' + Int2Str( CurLine ) + '->' +
        Macro.Items[ 0 ];
      LineIdx := 0;
      ToLineIdx := Macro.Count-1;
      CurLine := Macro.Objects[ 1 ];
      Inc( MacroLevel );
      CompileLines( TRUE, -1 );
    FINALLY
      EOF := FALSE;
      Lines := SaveSL;
      Compiler.CurFilePath := SavePath;
      CurLine := SaveCurLine;
      LineIdx := SaveLineIdx;
      ToLineIdx := SaveToLineIdx;
      LineBuffer := SaveLineBuffer;
      spos := @ LineBuffer[ 1 ];
      Inc( spos, Save_spos_offset );
      MacroLevel := SaveMacroLevel;
    END;
  end;
var i, n: Integer;
    patt, s, s2, paramname, expr: String;
    scanexpr: PChar;
    SaveParams: PStrListEx;
    params: PStrListEx;
begin
  Result := FALSE;
  if SkipLevel > 1 then Exit;
  if pos( '(', w ) <= 0 then //   
  for i := Compiler.MacroList.Count-1 downto 0 do
  begin
    if pos( '(', Compiler.MacroList.Items[ i ] ) > 0 then continue;
    if StrEq( Compiler.MacroList.Items[ i ], w ) then
    begin
      DoCallMacro( Pointer( Compiler.MacroList.Objects[ i ] ) );
      Result := TRUE;
      Exit;
    end;
  end
    else //   
  begin
    s := w;
    patt := '';
    while s <> '' do
    begin
      patt := patt + Parse( s, '(' ) + '(*)';
      Parse( s, ')' );
    end;
    for i := Compiler.MacroList.Count-1 downto 0 do
    begin
      if pos( '(', Compiler.MacroList.Items[ i ] ) <= 0 then continue;
      if StrSatisfy( Compiler.MacroList.Items[ i ], patt ) then
      begin
        params := NewStrListEx;
        TRY
          s := Compiler.MacroList.Items[ i ];
          s2 := w;
          while (s <> '') and (s2 <> '') do
          begin
            parse( s, '(' );
            parse( s2, '(' );
            if (s = '') or (s2 = '') then break;
            paramname := Trim( parse( s, ')' ) );
            if not IsIdentifier( paramname ) then
              Compiler.Error( 'Parameter name must be an identifier but ' + paramname + ' found' );
            expr := parse( s2, ')' );
            scanexpr := @ expr[ 1 ];
            n := Expression( scanexpr, '' );
            params.AddObject( paramname, n );
            s := Trim( s );
            s2 := Trim( s2 );
          end;
          if (s = '') <> (s2 = '') then
            Compiler.Error( 'Calling macro ' + w + ' with incorrect parameters count' );
          SaveParams := Compiler.MacroParams;
          Compiler.MacroParams := params;
          DoCallMacro( Pointer( Compiler.MacroList.Objects[ i ] ) );
          Compiler.MacroParams := SaveParams;
        FINALLY
          params.Free;
        END;
        Result := TRUE;
        Exit;
      end;
    end;
  end;
end;

procedure TProc.CompileCallProc(var w: String; var spos: PChar);
var proc_name, proc_name_1: String;
    i, n, npar, j0, ii: Integer;
    s: String;
begin
  if SkipLevel > 1 then Exit;
  if CompileCallMacro( w, spos ) then Exit;
  proc_name := w;
  npar := 0;
  i := pos( '<', w );
  if i > 1 then
  begin
    proc_name := Copy( w, 1, i-1 );
    npar := Str2Int( CopyEnd( w, i+1 ) );
    if npar > 3 then npar := 3;
    if npar < 0 then npar := 0;
  end;
  if UnitObj.UnitProcUsed.IndexOf_NoCase( proc_name ) < 0 then
    UnitObj.UnitProcUsed.Add( proc_name );
  proc_name_1 := proc_name;
  while StrReplace( proc_name_1, '.', '_' ) do;
  proc_name_1 := Copy( proc_name_1, 1, 30 );
  if not IsConcatOfIdentifiers( proc_name ) then
  begin
    if not CompileCallMacro( w, spos ) then
      Compiler.Error( 'Macro ' + w + ' not defined' );
    Exit;
  end;
  for j0 := 0 to 3 do
  begin
    for i := 0 to Compiler.ProcLists[ j0 ].Count do
    begin
      s := '';
      if i < Compiler.ProcLists[ j0 ].Count then
        s := Compiler.ProcLists[ j0 ].Items[ i ];
      if (i >= Compiler.ProcLists[ j0 ].Count) and (j0 = 3) or
         StrEq( s, proc_name ) then
      begin
        ii := i;
        if (ii >= Compiler.ProcLists[ j0 ].Count) then
        begin
          if Step = 1 then
          begin
            ii := Compiler.ProcLists[ npar ].Count;
            Compiler.ProcLists[ npar ].AddObject( proc_name, npar shl 24 or 1 )
          end
          else
            Compiler.Error( 'PROC ' + proc_name + ' not found' );
        end;
        n := Compiler.ProcLists[ npar ].Objects[ ii ];
        if (Step = 1) and (ProcMapped) then
        begin //   ,   
          if n shr 24 > 3 then
            n := n and $FFFFFF or (npar shl 24)
          else
          if n shr 24 <> npar then
            Compiler.Error( 'Number of parameters of ' + proc_name + '<' +
              Int2Str( npar ) + '> differ from defined earlier <' + Int2Str(
              n shr 24) + '>' );
          Inc( n );
          Compiler.ProcLists[ npar ].Objects[ ii ] := n;
        end;
        if (npar > 0) and (n and (1 shl 23) = 0) and
           not StrEq(Copy( w, 1, 6 ), 'PASCAL') or
           (npar = 0) and (ii > 63) then
        begin
          if ii <= 4095 then
          begin
            //OutByte( '$' + Int2Hex( $80 + (npar shl 4) + ii shr 8, 2 ) );
            //OutByte( '$' + Int2Hex( ii and $FF, 2 ) );
            OutWord( 'n_' + proc_name_1 );
          end
            else
          begin
            Compiler.Error( 'Procedure index > 4095: ' + proc_name );
            //todo:
          end;
        end
          else
        if i <= 63 then
        begin
          //OutByte( '$' + Int2Hex( $C0 + ii, 2 ) );
          OutByte( 'n_' + proc_name_1 );
        end
          else
        if i <= 16383 then
        begin
          //OutByte( '$' + Int2Hex( $80 + (ii shr 8), 2 ) );
          //OutByte( '$' + Int2Hex( i and $FF, 2 ) );
          OutWord( 'n_' + proc_name_1 );
        end
          else
        if i <= 65535 then
        begin
          s := 'CALLW';
          CompileCallProc( s, spos );
          //OutByte( '$' + Int2Hex( ii and $FF, 2 ) );
          //OutByte( '$' + Int2Hex( i shr 8, 2 ) );
          OutWord( 'n_' + proc_name_1 );
        end
          else
          Compiler.Error( 'Too large PROC number (' + Int2Str( ii ) + '): ' + proc_name );
        Exit;
      end;
    end;
  end;
end;

procedure TProc.CompileChars(w: String);
var i, L, p: Integer;
    n: String;
begin
  if SkipLevel > 1 then Exit;
  L := 0;
  i := 1;
  while i < Length( w ) do
  begin
    if w[ i ] = '#' then
    begin
      n := CopyEnd( Parse( w, ' ,'#9 ), 2 );
      OutByte( n ); inc( L );
      w := Trim( w );
      if (w <> '') and (w[ 1 ] = ',') then
        w := CopyEnd( w, 2 );
      i := 1; continue;
    end;
    if w[ i ] in [ '0'..'9' ] then
    begin
      n := Parse( w, ' ,'#9 );
      OutByte( n ); inc( L );
      w := Trim( w );
      if (w <> '') and (w[ 1 ] = ',') then
        w := CopyEnd( w, 2 );
      i := 1; continue;
    end;
    if w[ i ] = '''' then
    begin
      p := 0;
      while i < Length( w ) do
      begin
        inc( i );
        if w[ i ] = '''' then
        begin
          if (i < Length( w )) and (w[ i+1 ] = '''') then
            inc( i )
          else
          begin
            begin
              OutByte( Copy( w, 1, i ) );
              w := Trim( CopyEnd( w, i+1 ) );
              if (w <> '') and (w[ 1 ] = ',') then
                w := Trim( CopyEnd( w, 2 ) );
              i := 1;
            end;
            break;
          end;
        end;
        inc( L );
        inc( p );
        if (p >= 64) and not( (i < Length( w )) and (w[ i+1 ] = '''') ) then
        begin
          OutByte( Copy( w, 1, i ) + '''' );
          w := '''' + CopyEnd( w, i+1 );
          i := 1;
        end;
      end;
      if i = 1 then continue;
    end
      else Compiler.Error( 'Syntax string ' + w );
  end;
  if w <> '' then
    OutByte( w );
  Inc( Offset, L );
end;

procedure TProc.CompileElse;
var i, e: Integer;
begin
  if IFLevels.Count = 0 then
  begin
    Compiler.Error( 'ELSEIF without IF' );
    Exit;
  end;
  if IFLevels.Last = 'IF' then
  begin
    i := IFLevels.LastObj;
    e := FindLabel( #1'ENDIF' + Int2Str( i ), Offset + 1000 );
    OutJump( e, jAlways );
    if Step = 1 then
     UpdateLabel( #1'ELSE' + Int2Str( i ), Offset );
  end
    else
  if StrIn( IFLevels.Last, [ '.IF', '.ELSEIF' ] ) then
  begin
    if SkipLevel = 2 then
      SkipLevel := 1
    else if SkipLevel <= 1 then
      SkipLevel := 3;
  end
    else
    Compiler.Error( 'ELSE without IF' );
end;

procedure TProc.CompileElseif(var w: String);
var i: Integer;
    spos: PChar;
begin
  if IFLevels.Count = 0 then
  begin
    Compiler.Error( 'ELSEIF without IF' );
    Exit;
  end;
  if SkipLevel = 2 then
  begin
    i := pos( '(', w );
    spos := @ w[ i ];
    i := Expression( spos, '' );
    if i <> 0 then
      SkipLevel := 1;
  end
    else
  if SkipLevel = 1 then
    SkipLevel := 3;
  if StrIn( IFLevels.Last, [ '.IF', '.ELSEIF' ] ) then
    IFLevels.Items[ IFLevels.Count-1 ] := '.ELSEIF'
  else
    Compiler.Error( 'ELSEIF without IF' );
end;

procedure TProc.CompileEndif;
var i: Integer;
begin
  if IFLevels.Count = 0 then
    Compiler.Error( 'ENDIF without IF' )
  else
  if StrIn( IFLevels.Last, [ 'IF', 'ELSE' ] ) then
  begin
    i := IFLevels.LastObj;
    if Step = 1 then
      UpdateLabel( #1'ENDIF' + Int2Str( i ), Offset );
    IFLevels.Delete( IFLevels.Count-1 );
  end
    else
  if StrIn( IFLevels.Last, [ '.IF', '.ELSEIF', '.ELSE' ] ) then
  begin //   
    SkipLevel := IFLevels.LastObj;
    IFLevels.Delete( IFLevels.Count-1 );
  end;
end;

procedure TProc.CompileIF(var w: String);
var cond: TJumpCond;
    n: Integer;
    spos: PChar;
begin
  if StrEq( copy( w, 1, 6 ), 'IFDEF(' ) then
  begin
    IFLevels.AddObject( '.IF', SkipLevel );
    if SkipLevel >= 2 then
      Inc( SkipLevel, 2 )
    else
    begin
      Parse( w, '(' );
      w := Trim( Parse( w, ')' ) );
      if Compiler.ConditionalDefined( w ) then SkipLevel := 1
                                          else SkipLevel := 2;
    end;
  end
    else
  if w[ 3 ] = '(' then
  begin //  
    IFLevels.AddObject( '.IF', SkipLevel );
    if SkipLevel >= 2 then Inc( SkipLevel, 2 )
    else
    begin
      spos := @ w[ 3 ];
      n := Expression( spos, '' );
      if n = 0 then SkipLevel := 2
               else SkipLevel := 1;
    end;
  end
    else
  begin
    w := CopyEnd( w, 3 );
    if Copy( w, 1, 1 ) = '0' then
    begin
      cond := jIf0;
      w := CopyEnd( w, 2 );
    end
      else
    if Copy( w, 1, 1 ) = '1' then
    begin
      cond := jIf1;
      w := CopyEnd( w, 2 );
    end
      else
    begin
      cond := jAlways;
      Compiler.Error( 'Incorrect IF condition' );
    end;
    if (w <> '') and (w[ 1 ] = '(') then
    begin
      OutGoto( w, cond );
    end
      else
    begin //   IFcond ... ELSE ... ENDIF
      Inc( IFindex );
      IFLevels.AddObject( 'IF', IFindex );
      n := Offset + 1000;
      if Step = 1 then
      begin
        Labels.AddObject( #1'ELSE' + Int2Str( IFindex ), DWORD( -1 ) );
        Labels.AddObject( #1'ENDIF' + Int2Str( IFindex ), DWORD( -1 ) );
      end
        else
      begin
        n := FindLabel( #1'ELSE' + Int2Str( IFindex ), -1 );
        if n = -1 then
          n := FindLabel( #1'ENDIF' + Int2Str( IFindex ), -1 );
        if n < 0 then
          Compiler.Error( 'ENDIF not found' );
      end;
      OutJump( n, AlterCond( cond ) );
    end;
  end;
end;

procedure TProc.CompileLines(IsInlineMacro: Boolean; nPascalParams: Integer);
var spos: PChar;
    InProc: Boolean;
    InDebugLines: Boolean;
    ProcName: String;
    w: String;
    ProcIdx: Integer;
begin
  if nPascalParams >= 0 then
  begin
    {$IFDEF NT_ONLY}
    OutByte( '$' + Int2Hex( $E4 + nPascalParams, 2 ) );
    {$ELSE}
    //OutByte( '$CD' ); // INT n
    //OutByte( '$' + Int2Hex( $FC + nPascalParams, 2 ) );
    OutCode( ' CALL RunPCode' + Int2Str( nPascalParams ), 5 );
    {$ENDIF}
  end;
  ProcIdx := -1;
  ProcName := '';
  InProc := FALSE;
  InDebugLines := FALSE;
  EOF := FALSE;
  while not EOF do
  begin
    NextLine;
    if Trim( LineBuffer ) = '' then continue;
    if not InProc then
    begin //     -
      ProcName := Parse( LineBuffer, ' ' );
      LineBuffer := Trim( LineBuffer );
      if not StrIsStartingFrom( PChar( LineBuffer ), 'PROC' ) and (ProcName = '') then
      begin
        ProcName := Parse( LineBuffer, ' ' );
        LineBuffer := Trim( LineBuffer );
      end;
      if not StrIsStartingFrom( PChar( LineBuffer ), 'PROC' ) then
      begin
        Compiler.Error( 'PROC declaration not found' );
        Exit;
      end;
      inc( ProcIdx );
      if (ProcName = '') and (ProcIdx > 0) then
      begin
        Compiler.Error( 'Local PROC must have a name' );
        Exit;
      end;
      InProc := TRUE;
      IFLevels.Add( 'PROC' );
      if Compiler.MappedUnits <> nil then
      begin
        ProcMapped := Compiler.IsProcMapped( CurLine );
      end
        else
        ProcMapped := TRUE;
    end
      else
    begin
      spos := @ LineBuffer[ 1 ];
      //startpos := spos;
      while (spos <> nil) and (spos^ <> #0) and not EOF do
      begin
        SkipSp( spos );
        if (spos^ = '{') and (spos[ 1 ] = '$') and (spos[ 2 ] in ['R','r']) then
        begin
          w := SkipBrackets( spos );
          UnitObj.AddLineToTheEnd( w );
        end
          else
        if not InDebugLines and (spos^ = '{') then
        begin
          Inc( spos );
          InDebugLines := TRUE;
          if not Compiler.DebugMode then
            Inc( SkipLevel, 4 );
          continue;
        end
          else
        if (spos^ = '}') and InDebugLines then
        begin
          inc( spos );
          InDebugLines := FALSE;
          if not Compiler.DebugMode then
            Dec( SkipLevel, 4 );
          continue;
        end;
        if Step >= 2 then
          AddDebugInfo( LineIdx+1,
            DWORD( spos ) - DWORD( @ LineBuffer[ 1 ] ), Offset );
        w := SkipWord( spos, ': '#9, TRUE );
        if w = '' then continue;
        w := Trim( w );
        if StrEq( w, 'ENDP' ) then
        begin
          InProc := FALSE;
          if (IFLevels.Count > 0) and
             (IFLevels.Last <> 'PROC') then
          begin
            w := '';
            while (IFLevels.Count > 0) and
                  (IFLevels.Last <> 'PROC') do
            begin
              w := w + #13#10 + IFLevels.Last;
              IFLevels.Delete( IFLevels.Count-1 )
            end;
            Compiler.Error( 'Not balanced:'#13#10 + w );
            if (IFLevels.Count = 0) or (IFLevels.Last <> 'PROC') then
              Compiler.Error( 'ENDP without PROC' )
            else
              IFLevels.Delete( IFLevels.Count-1 );
          end
            else IFLevels.Delete( IFLevels.Count-1 );
          if IsInlineMacro then
          begin
            Exit; // ,  
          end;
          //todo:  INLINE-   -,  EXIT  RETURN
        end
          else
        begin //    -
          if w[ 1 ] = '#' then
          begin //  
            CompileBytes( w );
          end
            else
          if w[ 1 ] = '''' then
          begin //  
            CompileChars( w );
          end
            else
          if w[ Length( w ) ] = ':' then
          begin // 
            w := Copy( w, 1, Length( w ) - 1 );
            if (Step = 1) and (SkipLevel <= 1) then
              Labels.AddObject( w, Offset );
            if Copy( w, 1, 2 ) = '@@' then
              OutLabel( w );
          end
            else
          if StrEq( w, 'line' ) then
          begin
            OutWord( Int2Str( CurLine+1 ) );
          end
            else
          if StrIsStartingFrom( PChar( UpperCase( w ) ), 'GOTO') then
          begin
            OutGoto( w, jAlways );
          end
            else
          if StrEq(Copy( w, 1, 2 ), 'IF') and (Length( w ) > 2) then
          begin
            CompileIF( w );
          end
            else
          if StrIsStartingFrom( PChar( UpperCase( w ) ), 'ELSEIF(' ) then
          begin //    
            CompileElseif( w );
          end
            else
          if StrIsStartingFrom( PChar( UpperCase( w ) ), 'ELSE' ) then
          begin //     ,
                //     
            CompileElse;
          end
            else
          if StrIsStartingFrom( PChar( UpperCase( w ) ), 'ENDIF' ) then
          begin //    ,
                //    
            CompileEndif;
          end
            else
          if pos( '(', w ) > 0 then
          begin //    
            if not CompileCallMacro( w, spos ) and (SkipLevel <= 1) then
              Compiler.Error( 'Macro ' + w + ' not defined' );
          end
            else
          begin //    ,   -,  
            CompileCallProc( w, spos );
          end;
        end;
      end;
    end;
  end;
end;

destructor TProc.Destroy;
begin
  LineBuffer := '';
  IFLevels.Free;
  Labels.Free;
  inherited;
end;

function TProc.Expression(var spos: PChar; const op1: String): Integer;
var opd1, opd2: Integer;
    op2: String;
    op2start: PChar;
begin
  opd1 := Operand( spos );
  while TRUE do
  begin 
    SkipSp( spos );
    if spos^ in [ #0, ')', ']', '}' ] then
    begin
      Result := opd1;
      Exit;
    end;
    op2start := spos;
    if spos^ in [ '+', '-', '*', '/', '^', '&', '|', '=' ] then
    begin
      op2 := spos^;
      Inc( spos );
    end
      else
    if spos^ in [ '<' ] then
    begin
      if spos[ 1 ] in [ '<', '>', '=' ] then
      begin
        op2 := spos^ + spos[ 1 ];
        Inc( spos, 2 );
      end
        else
      begin
        op2 := spos^;
        Inc( spos );
      end;
    end
      else
    if spos^ in [ '>' ] then
    begin
      if spos[ 1 ] in [ '>', '=' ] then
      begin
        op2 := spos^ + spos[ 1 ];
        Inc( spos, 2 );
      end
        else
      begin
        op2 := spos^;
        Inc( spos );
      end;
    end
      else
      op2 := UpperCase( SkipWord( spos, ' }', FALSE ) );
    if not StrIn( op2, [ '+', '-', '*', '/', '^', '&', '|', '=', '<', '>', '<>',
      '<=', '>=', 'DIV', 'MOD', 'AND', 'OR', 'XOR', 'SHL', 'SHR', '<<', '>>' ] ) then
    begin
      Result := opd1;
      Compiler.Error( 'Invalid operation ' + op2 );
      Exit;
    end;
    if (op1 <> '') and (Priority( op1 ) >= Priority( op2 )) then
    begin
      spos := op2start;
      Result := opd1;
      Exit;
    end
      else
    if op2 <> '' then
    begin
      opd2 := Expression( spos, op2 );
      Result := Operation( op2, opd1, opd2 );
      opd1 := Result;
    end;
  end;
end;

function TProc.FindLabel(const s: String; DfltValue: Integer): Integer;
var i: Integer;
begin
  for i := 0 to Labels.Count-1 do
  begin
    if StrEq( s, Labels.Items[ i ] ) then
    begin
      Result := Labels.Objects[ i ]; Exit;
    end;
  end;
  Result := DfltValue;
  if Step >= 2 then
    Compiler.Error( 'Label ' + s + ' not found' );
end;

function TProc.HexNumber(const s: String): Integer;
var i: Integer;
begin
  Result := 0;
  for i := 1 to Length( s ) do
  begin
    if s[ i ] in [ '0'..'9' ] then
      Result := Result * 16 + Ord( s[ i ] ) - Ord( '0' )
    else if s[ i ] in [ 'a'..'z' ] then
      Result := Result * 16 + Ord( s[ i ] ) - Ord( 'a' ) + 10
    else if s[ i ] in [ 'A'..'Z' ] then
      Result := Result * 16 + Ord( s[ i ] ) - Ord( 'A' ) + 10
    else
    begin
      Compiler.Error( 'Hexadecimal number syntax: ' + s[ i ] );
      Exit;
    end;
  end;
end;

procedure TProc.Init;
begin
  inherited;
  IFLevels := NewStrListEx;
  Labels := NewStrListEx;
end;

procedure TProc.NewAsmLine;
begin
  if (MacroLevel > 0) or (SkipLevel > 0) then Exit;
  UnitObj.NeedNewAsmLine := TRUE;
end;

procedure TProc.NextLine;
var j, i, k: Integer;
    mparam: String;
begin
  inc( LineIdx );
  NewAsmLine;
  if LineIdx > ToLineIdx then
  begin
    LineBuffer := '';
    EOF := TRUE;
    Exit;
  end;
  LineBuffer := Lines.Items[ LineIdx ];
  CurLine := Lines.Objects[ LineIdx ];
  if InComment then
  begin
    j := pos( '*)', LineBuffer );
    if j > 0 then
    begin
      LineBuffer := CopyEnd( LineBuffer, j+2 );
      InComment := FALSE;
    end
      else
    begin
      LineBuffer := '';
      Exit;
    end;
  end;

  j := 1;
  while j <= Length( LineBuffer ) do
  begin
    if LineBuffer[ j ] = '''' then
    begin
      inc( j );
      while (j <= Length( LineBuffer )) and (LineBuffer[ j ] <> '''') do inc( j );
      inc( j );
      if j > Length( LineBuffer ) + 1 then
      begin
        Compiler.Error( 'Unterminated string' );
        break;
      end;
    end
      else
    if (j < Length( LineBuffer )) and
       (LineBuffer[ j ] = '/') and
       (LineBuffer[ j+1 ] = '/') then
    begin
      LineBuffer := Copy( LineBuffer, 1, j-1 );
      break;
    end
      else
    if (j < Length( LineBuffer )) and
       (LineBuffer[ j ] = '(') and
       (LineBuffer[ j+1 ] = '*') then
    begin
      k := pos( '*)', CopyEnd( LineBuffer, j+2 ) );
      if k > 0 then
      begin
        k := j+1 + k;
        LineBuffer := Copy( LineBuffer, 1, j-1 ) + ' ' +
                      CopyEnd( LineBuffer, k+2 );
        continue;
      end;
      LineBuffer := Copy( LineBuffer, 1, j-1 );
      InComment := TRUE;
      break;
    end
      else
    begin
      inc( j );
    end;
  end;

  if Compiler.MacroParams.Count > 0 then
  begin
    for j := Compiler.MacroParams.Count-1 downto 0 do
    begin
      mparam := '&' + Compiler.MacroParams.Items[ j ];
      for i := Length( LineBuffer )-Length(mparam)+1 downto 1 do
      begin
        if ( (i + Length( mparam ) >= Length( LineBuffer )) or
             not(LineBuffer[ Length( mparam ) + i ] in [ 'A'..'Z','a'..'z','_','0'..'9' ])
           ) and
           StrEq( Copy( LineBuffer, i, Length( mparam ) ), mparam ) then
          LineBuffer := Copy( LineBuffer, 1, i-1 ) +
            Int2Str( Compiler.MacroParams.Objects[ j ] ) +
            CopyEnd( LineBuffer, i+Length( mparam ) );
      end;
    end;
  end;
end;

function TProc.Number(const s: String): Integer;
var i: Integer;
begin
  Result := 0;
  if s = '' then
  begin
    Compiler.Error( 'Number syntax' );
    Exit;
  end;
  if s[1] = '$' then
  begin
    Result := HexNumber( CopyEnd( s, 2 ) );
  end
    else
  if s[ Length( s ) ] in [ 'h', 'H' ] then
  begin
    Result := HexNumber( Copy( s, 1, Length( s ) - 1 ) );
  end
    else
  begin
    for i := 1 to Length( s ) do
    begin
      if s[ i ] in [ '0'..'9' ] then
        Result := Result * 10 + Ord( s[ i ] ) - Ord( '0' )
      else
      begin
        Compiler.Error( 'Number syntax (not decimal digit ' + s[ i ] + ')' );
        Exit;
      end;
    end;
  end;
end;

function TProc.Operand(var spos: PChar): Integer;
  function DecodeOperand( const w: String ): Integer;
  var i: Integer;
  begin
    Result := 0;
    if w[ 1 ] = '-' then
      Result := -DecodeOperand( CopyEnd( w, 2 ) )
    else
    if w[ 1 ] = '~' then
      Result := not DecodeOperand( CopyEnd( w, 2 ) )
    else
    if w[ 1 ] = '+' then
      Result := DecodeOperand( CopyEnd( w, 2 ) )
    else
    if w[ 1 ] in [ '0'..'9', '$' ] then
    begin
      Result := Number( w );
    end
      else
    begin
      if w[ 1 ] in [ 'A'..'Z', 'a'..'z', '_' ] then
      begin
        for i := 0 to Compiler.MacroParams.Count-1 do
          if StrEq( w, Compiler.MacroParams.Items[ i ] ) then
          begin
            Result := Compiler.MacroParams.Objects[ i ];
            Exit;
          end;
      end;
      Compiler.Error( 'Operand ' + w + ' not recognized (must be a number or macro parameter)' );
    end;
  end;
var w: String;
begin
  Result := 0;
  SkipSp( spos );
  if spos^ = '(' then
  begin
    Inc( spos );
    Result := Expression( spos, '' );
    SkipSp( spos );
    if spos^ <> ')' then
      Compiler.Error( 'Waiting for '')''' )
    else
      Inc( spos );
    Exit;
  end;
  if spos^ = '-' then
  begin
    Inc( spos );
    SkipSp( spos );
    w := '-' + SkipWord( spos, '+-*/&^|<>=()[]{} '#9, FALSE );
  end
    else
    w := SkipWord( spos, '+-*/&^|<>=()[]{} '#9, FALSE );
  if Trim( w ) = '' then
  begin
    Compiler.Error( 'Waiting for operand' );
    Exit;
  end;
  Result := DecodeOperand( w );
end;

function TProc.Operation(const op: String; opd1, opd2: Integer): Integer;
  function Divide( x, y: Integer ): Integer;
  begin
    Result := 0;
    if y = 0 then
      Compiler.Error( 'Divide by 0' )
    else
      Result := x div y;
  end;
  function Modulo( x, y: Integer ): Integer;
  begin
    Result := 0;
    if y = 0 then
      Compiler.Error( 'Divide by 0' )
    else
      Result := x mod y;
  end;
begin
  Result := 0;
  if      op = '+'    then Result := opd1 + opd2
  else if op = '-'    then Result := opd1 - opd2
  else if op = '*'    then Result := opd1 * opd2
  else if op = '/'    then Result := Divide( opd1, opd2 )
  else if op = 'DIV'  then Result := Divide( opd1, opd2 )
  else if op = 'MOD'  then Result := Modulo( opd1, opd2 )
  else if op = '&'    then Result := opd1 and opd2
  else if op = 'AND'  then Result := opd1 and opd2
  else if op = '|'    then Result := opd1 or opd2
  else if op = 'OR'   then Result := opd1 or opd2
  else if op = '^'    then Result := opd1 xor opd2
  else if op = 'XOR'  then Result := opd1 xor opd2
  else if op = 'SHL'  then Result := opd1 shl opd2
  else if op = '<<'   then Result := opd1 shl opd2
  else if op = 'SHR'  then Result := opd1 shr opd2
  else if op = '>>'   then Result := opd1 shr opd2
  else if op = '<'    then Result := Integer( opd1 < opd2 )
  else if op = '<='   then Result := Integer( opd1 <= opd2 )
  else if op = '<>'   then Result := Integer( opd1 <> opd2 )
  else if op = '>'   then Result := Integer( opd1 > opd2 )
  else if op = '>='   then Result := Integer( opd1 >= opd2 )
  else if op = '='    then Result := Integer( opd1 = opd2 )
  else Compiler.Error( 'Invalid operation ' + op );
end;

procedure TProc.OutByte(const s: String);
var src: String;
begin
  if SkipLevel > 1 then Exit;
  if Step >= 2 then
  begin
    src := UnitObj.AsmCode.Last;
    if (Length( src ) < 72) and
       StrIsStartingFrom( PChar( src ), '  DB ' ) and not UnitObj.NeedNewAsmLine then
    begin
      src := src + ', ' + s;
      UnitObj.AsmCode.Items[ UnitObj.AsmCode.Count-1 ] := src;
    end
      else
    begin
      UnitObj.NeedNewAsmLine := FALSE;
      UnitObj.AsmCode.Add( '  DB ' + s );
    end;
  end;
  Inc( Offset );
end;

procedure TProc.OutCode(const s: String; Len: Integer);
begin
  if SkipLevel > 1 then Exit;
  if Step >= 2 then
  begin
      UnitObj.NeedNewAsmLine := TRUE;
      UnitObj.AsmCode.Add( s );
  end;
  Inc( Offset, Len );
end;

procedure TProc.OutDWord(const s: String);
var src: String;
begin
  if SkipLevel > 1 then Exit;
  if Step >= 2 then
  begin
    src := UnitObj.AsmCode.Last;
    if StrIsStartingFrom( PChar( src ), '  DD ' ) {and
       (Length( src ) + Length( s ) < 72)} and not UnitObj.NeedNewAsmLine then
    begin
      src := src + ', ' + s;
      UnitObj.AsmCode.Items[ UnitObj.AsmCode.Count-1 ] := src;
    end
      else
    begin
      UnitObj.NeedNewAsmLine := FALSE;
      UnitObj.AsmCode.Add( '  DD ' + s );
    end;
  end;
  Inc( Offset, 4 );
end;

procedure TProc.OutGoto(var w: String; cond: TJumpCond);
var n: Integer;
begin
  if SkipLevel > 1 then Exit;
  Parse( w, '(' );
  w := Trim( Parse( w, ')' ) );
  n := FindLabel( w, Offset+1000 );
  OutJump( n, cond );
end;

procedure TProc.OutJump(ToOffset: Integer; cond: TJumpCond);
var n: Integer;
    c: Integer;
begin
  if SkipLevel > 1 then Exit;
  n := (ToOffset - Offset - 2) and $1FFF;
  CASE cond OF
  jIf0: c := $40;
  jIf1: c := $60;
  else  c := $10;
  END;
  OutByte( '$' + Int2Hex( c + (n shr 8), 2 ) );
  OutByte( '$' + Int2Hex( (n and $FF), 2 ) );
end;

procedure TProc.OutLabel(const s: String);
begin
  if SkipLevel > 1 then Exit;
  if Step >= 2 then
  begin
    UnitObj.AsmCode.Add( s + ': ' );
  end;
end;

procedure TProc.OutQWord(const s: String);
var src: String;
begin
  if SkipLevel > 1 then Exit;
  if Step >= 2 then
  begin
    src := UnitObj.AsmCode.Last;
    if StrIsStartingFrom( PChar( src ), '  DQ ' ) {and
       (Length( src ) + Length( s ) < 72)} and not UnitObj.NeedNewAsmLine then
    begin
      src := src + ', ' + s;
      UnitObj.AsmCode.Items[ UnitObj.AsmCode.Count-1 ] := src;
    end
      else
    begin
      UnitObj.NeedNewAsmLine := FALSE;
      UnitObj.AsmCode.Add( '  DQ ' + s );
    end;
  end;
  Inc( Offset, 8 );
end;

procedure TProc.OutWord(const s: String);
var src: String;
begin
  if SkipLevel > 1 then Exit;
  if Step >= 2 then
  begin
    src := UnitObj.AsmCode.Last;
    if StrIsStartingFrom( PChar( src ), '  DW ' ) {and
       (Length( src ) + Length( s ) < 72)} and not UnitObj.NeedNewAsmLine then
    begin
      src := src + ', ' + s;
      UnitObj.AsmCode.Items[ UnitObj.AsmCode.Count-1 ] := src;
    end
      else
    begin
      UnitObj.NeedNewAsmLine := FALSE;
      UnitObj.AsmCode.Add( '  DW ' + s );
    end;
  end;
  Inc( Offset, 2 );
end;

function TProc.Priority(const op: String): Integer;
const Ops = 'OR | 1 AND & XOR ^ 2 NOT ~ 3 <= < <> > >= = 4 SHL << SHR >> 5 ' +
  '+ - 6 * / DIV MOD 7';
var i: Integer;
begin
  i := pos( op, Ops );
  if i > 0 then
  for i := i to Length( Ops ) do
  begin
    if Ops[ i ] in [ '0'..'9' ] then
    begin
      Result := Ord( Ops[ i ] ) - Ord( '0' );
      Exit;
    end;
  end;
  Result := -1;
end;

function TProc.SkipBrackets(var spos: PChar): String;
var OpenBracket, CloseBracket: Char;
    startpos: PChar;
    S1: String;
begin
  Result := '';
  startpos := spos;
  OpenBracket := spos^;
  CASE OpenBracket OF
  '(': CloseBracket := ')';
  '[': CloseBracket := ']';
  '{': CloseBracket := '}';
  '<': CloseBracket := '>';
  else Compiler.Error( 'Incorrect open bracket: ' + OpenBracket );
       Inc( spos );
       Exit;
  END;
  Inc( spos );
  while (spos <> nil) and (spos^ <> CloseBracket) and not EOF do
  begin
    if spos^ in [ '(', '[', '{' ] then
    begin
      SetString( S1, startpos, DWORD( spos ) - DWORD( startpos ) );
      Result := Result + S1;
      Result := Result + SkipBrackets( spos );
      startpos := spos;
    end
      else
    if spos^ = '''' then
    begin
      SetString( S1, startpos, DWORD( spos ) - DWORD( startpos ) );
      Result := Result + S1;
      Result := Result + SkipString( spos );
      startpos := spos;
    end
      else
    begin
      Inc( spos );
      if spos^ = #0 then
      begin
        SetString( S1, startpos, DWORD( spos ) - DWORD( startpos ) );
        Result := Result + S1;
        NextLine;
        spos := nil;
        if LineBuffer <> '' then
          spos := @ LineBuffer[ 1 ];
        startpos := spos;
      end;
    end;
  end;
  if spos <> nil then
  begin
    if spos^ = CloseBracket then
      Inc( spos );
    SetString( S1, startpos, DWORD( spos ) - DWORD( startpos ) );
  end;
  Result := Result + S1;
end;

procedure TProc.SkipSp(var spos: PChar);
begin
  while (spos^ <> #0) and (spos^ <= ' ') do inc( spos );
end;

function TProc.SkipString(var spos: PChar): String;
var //Buf: String;
    //i: Integer;
    startpos: PChar;
begin
  if spos^ <> '''' then
    Compiler.Error( 'Incorrect string ' + spos );
  startpos := spos;
  Inc( spos );
  //Buf := spos;
  //i := 0;
  while spos^ <> #0 do
  begin
    if spos^ = '''' then
    begin
      Inc( spos );
      if spos^ = '''' then
      begin
        inc( spos );
        //inc( i );
        //Buf[ i ] := '''';
      end
        else
        break;
    end
      else
    begin
      //inc( i );
      //Buf[ i ] := spos^;
      inc( spos );
      if spos^ = #0 then
        Compiler.Error( 'Unterminated string' );
    end;
  end;
  //inc( i );
  //Buf[ i ] := #0;
  //Result := PChar( Buf );
  SetString( Result, startpos, DWORD( spos ) - DWORD( startpos ) );
end;

function TProc.SkipWord(var spos: PChar; const Delimiters: String;
  IncludingDelimiter: Boolean): String;
var s1: String;
    startpos: PChar;
begin
  Result := '';
  startpos := spos;
  while (spos <> nil) and (spos^ > ' ') do
  begin
    if spos^ in [ '(', '[', '{' ] then
    begin
      SetString( s1, startpos, DWORD( spos ) - DWORD( startpos ) );
      Result := Result + s1;
      Result := Result + SkipBrackets( spos );
      startpos := spos;
    end
      else
    if spos^ in [ '}' ] then
    begin
      break;
    end
      else
    if pos( spos^, Delimiters ) > 0 then
    begin
      if IncludingDelimiter then
        inc( spos );
      SetString( s1, startpos, DWORD( spos ) - DWORD( startpos ) );
      Result := Result + s1;
      Exit;
    end
      else
    if spos^ = '''' then
    begin
      SetString( s1, startpos, DWORD( spos ) - DWORD( startpos ) );
      Result := Result + s1;
      Result := Result + SkipString( spos );
      startpos := spos;
    end
      else
      inc( spos );
  end;
  SetString( s1, startpos, DWORD( spos ) - DWORD( startpos ) );
  Result := Result + s1;
end;

procedure TProc.UpdateLabel(const s: String; NewValue: Integer);
var i: Integer;
begin
  for i := 0 to Labels.Count-1 do
  begin
    if StrEq( s, Labels.Items[ i ] ) then
    begin
      Labels.Objects[ i ] := NewValue; Exit;
    end;
  end;
  if Step >= 2 then
    Compiler.Error( 'Label ' + s + ' not found' );
end;

end.
